# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Functionality for styling plots"""
import os

import numpy as np

from matplotlib.patches import Polygon
from scipy.linalg import norm
from scipy.spatial import ConvexHull
from neurom._compat import map

# needed so that projection='3d' works with fig.add_subplot
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import


plt = None  # refer to _get_plt()


def _get_plt():
    '''wrapper to avoid loading matplotlib.pyplot before someone has a chance to set the backend'''
    global plt  # pylint: disable=global-statement
    import matplotlib.pyplot
    plt = matplotlib.pyplot


def dict_if_none(arg):
    '''return an empty dict if arg is None'''
    return arg if arg is not None else {}


def figure_naming(pretitle='', posttitle='', prefile='', postfile=''):
    """
    Helper function to define the strings that handle pre-post conventions
    for viewing - plotting title and saving options.

    Args:
        pretitle(str): String to include before the general title of the figure.
        posttitle(str): String to include after the general title of the figure.
        prefile(str): String to include before the general filename of the figure.
        postfile(str): String to include after the general filename of the figure.

    Returns:
        str: String to include in the figure name and title, in a suitable form.
    """
    if pretitle:
        pretitle = "%s -- " % pretitle

    if posttitle:
        posttitle = " -- %s" % posttitle

    if prefile:
        prefile = "%s_" % prefile

    if postfile:
        postfile = "_%s" % postfile

    return pretitle, posttitle, prefile, postfile


def get_figure(new_fig=True, subplot='111', params=None):
    """
    Function to be used for viewing - plotting,
    to initialize the matplotlib figure - axes.

    Args:
        new_fig(bool): Defines if a new figure will be created, if false current figure is used
        subplot (tuple or matplolib subplot specifier string): Create axes with these parameters
        params (dict): extra options passed to add_subplot()

    Returns:
        Matplotlib Figure and Axes
    """
    _get_plt()

    if new_fig:
        fig = plt.figure()
    else:
        fig = plt.gcf()

    params = dict_if_none(params)

    if isinstance(subplot, (tuple, list)):
        ax = fig.add_subplot(*subplot, **params)
    else:
        ax = fig.add_subplot(subplot, **params)

    return fig, ax


def save_plot(fig, prefile='', postfile='', output_path='./', output_name='Figure',
              output_format='png', dpi=300, transparent=False, **_):
    """Generates a figure file in the selected directory.

    Args:
        fig: matplotlib figure
        prefile(str): Include before the general filename of the figure
        postfile(str): Included after the general filename of the figure
        output_path(str): Define the path to the output directory
        output_name(str): String to define the name of the output figure
        output_format(str): String to define the format of the output figure
        dpi(int): Define the DPI (Dots per Inch) of the figure
        transparent(bool): If True the saved figure will have a transparent background
    """
    if not os.path.exists(output_path):
        os.makedirs(output_path)  # Make output directory if non-exsiting

    output = os.path.join(output_path,
                          prefile + output_name + postfile + "." + output_format)

    fig.savefig(output, dpi=dpi, transparent=transparent)


def plot_style(fig, ax,  # pylint: disable=too-many-arguments, too-many-locals
               # plot_title
               pretitle='',
               title='Figure',
               posttitle='',
               title_fontsize=14,
               title_arg=None,
               # plot_labels
               label_fontsize=14,
               xlabel=None,
               xlabel_arg=None,
               ylabel=None,
               ylabel_arg=None,
               zlabel=None,
               zlabel_arg=None,
               # plot_ticks
               tick_fontsize=12,
               xticks=None,
               xticks_args=None,
               yticks=None,
               yticks_args=None,
               zticks=None,
               zticks_args=None,
               # update_plot_limits
               white_space=30,
               # plot_legend
               no_legend=True,
               legend_arg=None,
               # internal
               no_axes=False,
               aspect_ratio='equal',
               tight=False,
               **_):
    """Set the basic options of a matplotlib figure, to be used by viewing - plotting functions

    Args:
        fig(matplotlib figure): figure
        ax(matplotlib axes, belonging to `fig`): axes

        pretitle(str): String to include before the general title of the figure
        posttitle (str): String to include after the general title of the figure
        title (str): Set the title for the figure
        title_fontsize (int): Defines the size of the title's font
        title_arg (dict): Addition arguments for matplotlib.title() call

        label_fontsize(int): Size of the labels' font
        xlabel(str): The xlabel for the figure
        xlabel_arg(dict):  Passsed into matplotlib as xlabel arguments
        ylabel(str): The ylabel for the figure
        ylabel_arg(dict):  Passsed into matplotlib as ylabel arguments
        zlabel(str): The zlabel for the figure
        zlabel_arg(dict):  Passsed into matplotlib as zlabel arguments

        tick_fontsize (int): Defines the size of the ticks' font
        xticks([list of ticks]): Defines the values of x ticks in the figure
        xticks_args(dict):  Passsed into matplotlib as xticks arguments
        yticks([list of ticks]): Defines the values of y ticks in the figure
        yticks_args(dict):  Passsed into matplotlib as yticks arguments
        zticks([list of ticks]): Defines the values of z ticks in the figure
        zticks_args(dict):  Passsed into matplotlib as zticks arguments

        white_space(float): whitespace added to surround the tight limit of the data

        no_legend (bool): Defines the presence of a legend in the figure
        legend_arg (dict): Addition arguments for matplotlib.legend() call

        no_axes(bool): If True the labels and the frame will be set off
        aspect_ratio(str): Sets aspect ratio of the figure, according to matplotlib aspect_ratio
        tight(bool): If True the tight layout of matplotlib will be activated

    Returns:
        Matplotlib figure, matplotlib axes
    """
    plot_title(ax, pretitle, title, posttitle, title_fontsize, title_arg)
    plot_labels(ax, label_fontsize, xlabel, xlabel_arg, ylabel, ylabel_arg, zlabel, zlabel_arg)
    plot_ticks(ax, tick_fontsize, xticks, xticks_args, yticks, yticks_args, zticks, zticks_args)
    update_plot_limits(ax, white_space)
    plot_legend(ax, no_legend, legend_arg)

    if no_axes:
        ax.set_frame_on(False)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

    ax.set_aspect(aspect_ratio)

    if tight:
        fig.set_tight_layout(True)


def plot_title(ax, pretitle='', title='Figure', posttitle='', title_fontsize=14, title_arg=None):
    """Set title options of a matplotlib plot

    Args:
        ax: matplotlib axes
        pretitle(str): String to include before the general title of the figure
        posttitle (str): String to include after the general title of the figure
        title (str): Set the title for the figure
        title_fontsize (int): Defines the size of the title's font
        title_arg (dict): Addition arguments for matplotlib.title() call
    """
    current_title = ax.get_title()

    if not current_title:
        current_title = pretitle + title + posttitle

    title_arg = dict_if_none(title_arg)

    ax.set_title(current_title, fontsize=title_fontsize, **title_arg)


def plot_labels(ax, label_fontsize=14,
                xlabel=None, xlabel_arg=None,
                ylabel=None, ylabel_arg=None,
                zlabel=None, zlabel_arg=None):
    """Sets the labels options of a matplotlib plot

    Args:
        ax: matplotlib axes
        label_fontsize(int): Size of the labels' font
        xlabel(str): The xlabel for the figure
        xlabel_arg(dict):  Passsed into matplotlib as xlabel arguments
        ylabel(str): The ylabel for the figure
        ylabel_arg(dict):  Passsed into matplotlib as ylabel arguments
        zlabel(str): The zlabel for the figure
        zlabel_arg(dict):  Passsed into matplotlib as zlabel arguments
    """
    xlabel = xlabel if xlabel is not None else ax.get_xlabel() or 'X'
    ylabel = ylabel if ylabel is not None else ax.get_ylabel() or 'Y'

    xlabel_arg = dict_if_none(xlabel_arg)
    ylabel_arg = dict_if_none(ylabel_arg)

    ax.set_xlabel(xlabel, fontsize=label_fontsize, **xlabel_arg)
    ax.set_ylabel(ylabel, fontsize=label_fontsize, **ylabel_arg)

    if hasattr(ax, 'zaxis'):
        zlabel = zlabel if zlabel is not None else ax.get_zlabel() or 'Z'
        zlabel_arg = dict_if_none(zlabel_arg)
        ax.set_zlabel(zlabel, fontsize=label_fontsize, **zlabel_arg)


def plot_ticks(ax, tick_fontsize=12,
               xticks=None, xticks_args=None,
               yticks=None, yticks_args=None,
               zticks=None, zticks_args=None):
    """Function that defines the labels options of a matplotlib plot.

    Args:
        ax: matplotlib axes
        tick_fontsize (int): Defines the size of the ticks' font
        xticks([list of ticks]): Defines the values of x ticks in the figure
        xticks_arg(dict):  Passsed into matplotlib as xticks arguments
        yticks([list of ticks]): Defines the values of y ticks in the figure
        yticks_arg(dict):  Passsed into matplotlib as yticks arguments
        zticks([list of ticks]): Defines the values of z ticks in the figure
        zticks_arg(dict):  Passsed into matplotlib as zticks arguments
    """
    if xticks is not None:
        ax.set_xticks(xticks)
        xticks_args = dict_if_none(xticks_args)
        ax.xaxis.set_tick_params(labelsize=tick_fontsize, **xticks_args)

    if yticks is not None:
        ax.set_yticks(yticks)
        yticks_args = dict_if_none(yticks_args)
        ax.yaxis.set_tick_params(labelsize=tick_fontsize, **yticks_args)

    if zticks is not None:
        ax.set_zticks(zticks)
        zticks_args = dict_if_none(zticks_args)
        ax.zaxis.set_tick_params(labelsize=tick_fontsize, **zticks_args)


def update_plot_limits(ax, white_space):
    """Sets the limit options of a matplotlib plot.

    Args:
        ax: matplotlib axes
        white_space(float): whitespace added to surround the tight limit of the data

    Note: This relies on ax.dataLim (in 2d) and ax.[xy, zz]_dataLim being set in 3d
    """

    if hasattr(ax, 'zz_dataLim'):
        bounds = ax.xy_dataLim.bounds
        ax.set_xlim(bounds[0] - white_space, bounds[0] + bounds[2] + white_space)
        ax.set_ylim(bounds[1] - white_space, bounds[1] + bounds[3] + white_space)

        bounds = ax.zz_dataLim.bounds
        ax.set_zlim(bounds[0] - white_space, bounds[0] + bounds[2] + white_space)
    else:
        bounds = ax.dataLim.bounds
        assert not any(map(np.isinf, bounds)), 'Cannot set bounds if dataLim has infinite elements'
        ax.set_xlim(bounds[0] - white_space, bounds[0] + bounds[2] + white_space)
        ax.set_ylim(bounds[1] - white_space, bounds[1] + bounds[3] + white_space)


def plot_legend(ax, no_legend=True, legend_arg=None):
    """
    Function that defines the legend options
    of a matplotlib plot.

    Args:
        ax: matplotlib axes
        no_legend (bool): Defines the presence of a legend in the figure
        legend_arg (dict): Addition arguments for matplotlib.legend() call
    """
    legend_arg = dict_if_none(legend_arg)

    if not no_legend:
        ax.legend(**legend_arg)


_LINSPACE_COUNT = 300


def _get_normals(v):
    '''get two vectors that form a basis w/ v

    Note: returned vectors are unit
    '''
    not_v = np.array([1, 0, 0])
    if np.all(np.abs(v) == not_v):
        not_v = np.array([0, 1, 0])
    n1 = np.cross(v, not_v)
    n1 /= norm(n1)
    n2 = np.cross(v, n1)
    return n1, n2


def generate_cylindrical_points(start, end, start_radius, end_radius,
                                linspace_count=_LINSPACE_COUNT):
    '''Generate a 3d mesh of a cylinder with start and end points, and varying radius

    Based on: http://stackoverflow.com/a/32383775
    '''
    v = end - start
    length = norm(v)
    v = v / length
    n1, n2 = _get_normals(v)

    # pylint: disable=unbalanced-tuple-unpacking
    l, theta = np.meshgrid(np.linspace(0, length, linspace_count),
                           np.linspace(0, 2 * np.pi, linspace_count))

    radii = np.linspace(start_radius, end_radius, linspace_count)
    rsin = np.multiply(radii, np.sin(theta))
    rcos = np.multiply(radii, np.cos(theta))

    return np.array([start[i] +
                     v[i] * l +
                     n1[i] * rsin + n2[i] * rcos
                     for i in range(3)])


def project_cylinder_onto_2d(ax, plane,
                             start, end, start_radius, end_radius,
                             color='black', alpha=1.):
    '''take cylinder defined by start/end, and project it onto the plane

    Args:
        ax: matplotlib axes
        plane(tuple of int): where x, y, z = 0, 1, 2, so (0, 1) is the xy axis
        start(np.array): start coordinates
        end(np.array): end coordinates
        start_radius(float): start radius
        end_radius(float): end radius
        color: matplotlib color
        alpha(float): alpha value

    Note: There are probably more efficient ways of doing this: here the
    3d outline is calculated, the non-used plane coordinates are dropped, a
    tight convex hull is found, and that is used for a filled polygon
    '''
    points = generate_cylindrical_points(start, end, start_radius, end_radius, 10)
    points = np.vstack([points[plane[0]].ravel(),
                        points[plane[1]].ravel()])
    points = points.T
    hull = ConvexHull(points)
    ax.add_patch(Polygon(points[hull.vertices], fill=True, color=color, alpha=alpha))


def plot_cylinder(ax, start, end, start_radius, end_radius,
                  color='black', alpha=1., linspace_count=_LINSPACE_COUNT):
    '''plot a 3d cylinder'''
    assert not np.all(start == end), 'Cylinder must have length'
    x, y, z = generate_cylindrical_points(start, end, start_radius, end_radius,
                                          linspace_count=linspace_count)
    ax.plot_surface(x, y, z, color=color, alpha=alpha)


def plot_sphere(ax, center, radius, color='black', alpha=1., linspace_count=_LINSPACE_COUNT):
    """ Plots a 3d sphere, given the center and the radius.  """

    u = np.linspace(0, 2 * np.pi, linspace_count)
    v = np.linspace(0, np.pi, linspace_count)
    sin_v = np.sin(v)
    x = center[0] + radius * np.outer(np.cos(u), sin_v)
    y = center[1] + radius * np.outer(np.sin(u), sin_v)
    z = center[2] + radius * np.outer(np.ones_like(u), np.cos(v))

    ax.plot_surface(x, y, z, linewidth=0.0, color=color, alpha=alpha)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Neurite functions'''

from itertools import chain

import numpy as np

from neurom import morphmath
from neurom.core import Tree, iter_neurites, iter_sections, NeuriteType
from neurom.core.dataformat import COLS
from neurom.core.types import tree_type_checker as is_type
from neurom.fst import _bifurcationfunc
from neurom.fst import _neuronfunc
from neurom.fst import sectionfunc
from neurom.geom import convex_hull


def total_length(nrn_pop, neurite_type=NeuriteType.all):
    '''Get the total length of all sections in the group of neurons or neurites'''
    nrns = _neuronfunc.neuron_population(nrn_pop)
    return list(sum(section_lengths(n, neurite_type=neurite_type)) for n in nrns)


def n_segments(neurites, neurite_type=NeuriteType.all):
    '''Number of segments in a collection of neurites'''
    return sum(len(s.points) - 1
               for s in iter_sections(neurites, neurite_filter=is_type(neurite_type)))


def n_neurites(neurites, neurite_type=NeuriteType.all):
    '''Number of neurites in a collection of neurites'''
    return sum(1 for _ in iter_neurites(neurites, filt=is_type(neurite_type)))


def n_sections(neurites, neurite_type=NeuriteType.all, iterator_type=Tree.ipreorder):
    '''Number of sections in a collection of neurites'''
    return sum(1 for _ in iter_sections(neurites,
                                        iterator_type=iterator_type,
                                        neurite_filter=is_type(neurite_type)))


def n_bifurcation_points(neurites, neurite_type=NeuriteType.all):
    '''number of bifurcation points in a collection of neurites'''
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.ibifurcation_point)


def n_forking_points(neurites, neurite_type=NeuriteType.all):
    '''number of forking points in a collection of neurites'''
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.iforking_point)


def n_leaves(neurites, neurite_type=NeuriteType.all):
    '''number of leaves points in a collection of neurites'''
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.ileaf)


def map_sections(fun, neurites, neurite_type=NeuriteType.all, iterator_type=Tree.ipreorder):
    '''Map `fun` to all the sections in a collection of neurites'''
    return map(fun, iter_sections(neurites,
                                  iterator_type=iterator_type,
                                  neurite_filter=is_type(neurite_type)))


def _section_length(section):
    '''get section length of `section`'''
    return morphmath.section_length(section.points)


def section_lengths(neurites, neurite_type=NeuriteType.all):
    '''section lengths in a collection of neurites'''
    return map_sections(_section_length, neurites, neurite_type=neurite_type)


def section_term_lengths(neurites, neurite_type=NeuriteType.all):
    '''Termination section lengths in a collection of neurites'''
    return map_sections(_section_length, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ileaf)


def section_bif_lengths(neurites, neurite_type=NeuriteType.all):
    '''Bifurcation section lengths in a collection of neurites'''
    return map_sections(_section_length, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def section_branch_orders(neurites, neurite_type=NeuriteType.all):
    '''section branch orders in a collection of neurites'''
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type)


def section_bif_branch_orders(neurites, neurite_type=NeuriteType.all):
    '''Bifurcation section branch orders in a collection of neurites'''
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def section_term_branch_orders(neurites, neurite_type=NeuriteType.all):
    '''Termination section branch orders in a collection of neurites'''
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ileaf)


def section_path_lengths(neurites, neurite_type=NeuriteType.all):
    '''Path lengths of a collection of neurites '''
    # Calculates and stores the section lengths in one pass,
    # then queries the lengths in the path length iterations.
    # This avoids repeatedly calculating the lengths of the
    # same sections.
    dist = {}
    neurite_filter = is_type(neurite_type)

    for s in iter_sections(neurites, neurite_filter=neurite_filter):
        dist[s] = s.length

    def pl2(node):
        '''Calculate the path length using cached section lengths'''
        return sum(dist[n] for n in node.iupstream())

    return map_sections(pl2, neurites, neurite_type=neurite_type)


def map_neurons(fun, neurites, neurite_type):
    '''Map `fun` to all the neurites in a single or collection of neurons'''
    nrns = _neuronfunc.neuron_population(neurites)
    return [fun(n, neurite_type=neurite_type) for n in nrns]


def number_of_sections(neurites, neurite_type=NeuriteType.all):
    '''Number of sections in a collection of neurites'''
    return map_neurons(n_sections, neurites, neurite_type)


def number_of_neurites(neurites, neurite_type=NeuriteType.all):
    '''Number of neurites in a collection of neurites'''
    return map_neurons(n_neurites, neurites, neurite_type)


def number_of_bifurcations(neurites, neurite_type=NeuriteType.all):
    '''number of bifurcation points in a collection of neurites'''
    return map_neurons(n_bifurcation_points, neurites, neurite_type)


def number_of_forking_points(neurites, neurite_type=NeuriteType.all):
    '''number of forking points in a collection of neurites'''
    return map_neurons(n_forking_points, neurites, neurite_type)


def number_of_terminations(neurites, neurite_type=NeuriteType.all):
    '''number of leaves points in a collection of neurites'''
    return map_neurons(n_leaves, neurites, neurite_type)


def number_of_segments(neurites, neurite_type=NeuriteType.all):
    '''Number of sections in a collection of neurites'''
    return map_neurons(n_segments, neurites, neurite_type)


def map_segments(func, neurites, neurite_type):
    ''' Map `func` to all the segments in a collection of neurites

        `func` accepts a section and returns list of values corresponding to each segment.
    '''
    neurite_filter = is_type(neurite_type)
    return [
        s for ss in iter_sections(neurites, neurite_filter=neurite_filter) for s in func(ss)
    ]


def segment_lengths(neurites, neurite_type=NeuriteType.all):
    '''Lengths of the segments in a collection of neurites'''
    def _seg_len(sec):
        '''list of segment lengths of a section'''
        return np.linalg.norm(np.diff(sec.points[:, COLS.XYZ], axis=0), axis=1)

    return map_segments(_seg_len, neurites, neurite_type)


def segment_volumes(neurites, neurite_type=NeuriteType.all):
    '''Volumes of the segments in a collection of neurites'''
    def _func(sec):
        '''list of segment volumes of a section'''
        return [morphmath.segment_volume(seg) for seg in zip(sec.points[:-1], sec.points[1:])]

    return map_segments(_func, neurites, neurite_type)


def segment_radii(neurites, neurite_type=NeuriteType.all):
    '''arithmetic mean of the radii of the points in segments in a collection of neurites'''
    def _seg_radii(sec):
        '''vectorized mean radii'''
        pts = sec.points[:, COLS.R]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return map_segments(_seg_radii, neurites, neurite_type)


def segment_taper_rates(neurites, neurite_type=NeuriteType.all):
    '''taper rates of the segments in a collection of neurites

    The taper rate is defined as the absolute radii differences divided by length of the section
    '''
    def _seg_taper_rates(sec):
        '''vectorized taper rates'''
        pts = sec.points[:, COLS.XYZR]
        diff = np.diff(pts, axis=0)
        distance = np.linalg.norm(diff[:, COLS.XYZ], axis=1)
        return np.divide(2 * np.abs(diff[:, COLS.R]), distance)

    return map_segments(_seg_taper_rates, neurites, neurite_type)


def segment_meander_angles(neurites, neurite_type=NeuriteType.all):
    '''Inter-segment opening angles in a section'''
    return list(chain.from_iterable(map_sections(
        sectionfunc.section_meander_angles, neurites, neurite_type)))


def segment_midpoints(neurites, neurite_type=NeuriteType.all):
    '''Return a list of segment mid-points in a collection of neurites'''
    def _seg_midpoint(sec):
        '''Return the mid-points of segments in a section'''
        pts = sec.points[:, COLS.XYZ]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return map_segments(_seg_midpoint, neurites, neurite_type)


def segment_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None):
    '''Lengths of the segments in a collection of neurites'''
    def _seg_rd(sec, pos):
        '''list of radial distances of all segments of a section'''
        mid_pts = np.divide(np.add(sec.points[:-1], sec.points[1:])[:, :3], 2.0)
        return np.sqrt([morphmath.point_dist2(p, pos) for p in mid_pts])

    dist = []
    for n in iter_neurites(neurites, filt=is_type(neurite_type)):
        pos = n.root_node.points[0] if origin is None else origin
        dist.extend([s for ss in n.iter_sections() for s in _seg_rd(ss, pos)])

    return dist


def local_bifurcation_angles(neurites, neurite_type=NeuriteType.all):
    '''Get a list of local bifurcation angles in a collection of neurites'''
    return map_sections(_bifurcationfunc.local_bifurcation_angle,
                        neurites,
                        neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def remote_bifurcation_angles(neurites, neurite_type=NeuriteType.all):
    '''Get a list of remote bifurcation angles in a collection of neurites'''
    return map_sections(_bifurcationfunc.remote_bifurcation_angle,
                        neurites,
                        neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def bifurcation_partitions(neurites, neurite_type=NeuriteType.all):
    '''Partition at bifurcation points of a collection of neurites'''
    return map(_bifurcationfunc.bifurcation_partition,
               iter_sections(neurites,
                             iterator_type=Tree.ibifurcation_point,
                             neurite_filter=is_type(neurite_type)))


def partition_asymmetries(neurites, neurite_type=NeuriteType.all):
    '''Partition asymmetry at bifurcation points of a collection of neurites'''
    return map(_bifurcationfunc.partition_asymmetry,
               iter_sections(neurites,
                             iterator_type=Tree.ibifurcation_point,
                             neurite_filter=is_type(neurite_type)))


def section_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None):
    '''Remote bifurcation angles in a collection of neurites'''
    dist = []
    for n in iter_neurites(neurites, filt=is_type(neurite_type)):
        pos = n.root_node.points[0] if origin is None else origin
        dist.extend([sectionfunc.section_radial_distance(s, pos)
                     for s in n.iter_sections()])

    return dist


def number_of_sections_per_neurite(neurites, neurite_type=NeuriteType.all):
    '''Get the number of sections per neurite in a collection of neurites'''
    return list(sum(1 for _ in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def total_length_per_neurite(neurites, neurite_type=NeuriteType.all):
    '''Get the path length per neurite in a collection'''
    return list(sum(s.length for s in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def terminal_path_lengths_per_neurite(neurites, neurite_type=NeuriteType.all):
    '''Get the path lengths to each terminal point per neurite in a collection'''
    return list(sectionfunc.section_path_length(s)
                for n in iter_neurites(neurites, filt=is_type(neurite_type))
                for s in iter_sections(n, iterator_type=Tree.ileaf))


def total_volume_per_neurite(neurites, neurite_type=NeuriteType.all):
    '''Get the volume per neurite in a collection'''
    return list(sum(s.volume for s in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def neurite_volume_density(neurites, neurite_type=NeuriteType.all):
    '''Get the volume density per neurite

    The volume density is defined as the ratio of the neurite volume and
    the volume of the neurite's enclosung convex hull
    '''
    def vol_density(neurite):
        '''volume density of a single neurite'''
        return neurite.volume / convex_hull(neurite).volume

    return list(vol_density(n)
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def section_volumes(neurites, neurite_type=NeuriteType.all):
    '''section volumes in a collection of neurites'''
    return map_sections(sectionfunc.section_volume, neurites, neurite_type=neurite_type)


def section_areas(neurites, neurite_type=NeuriteType.all):
    '''section areas in a collection of neurites'''
    return map_sections(sectionfunc.section_area, neurites, neurite_type=neurite_type)


def section_tortuosity(neurites, neurite_type=NeuriteType.all):
    '''section tortuosities in a collection of neurites'''
    return map_sections(sectionfunc.section_tortuosity, neurites, neurite_type=neurite_type)


def principal_direction_extents(neurites, neurite_type=NeuriteType.all, direction=0):
    '''Principal direction extent of neurites in neurons'''
    def _pde(neurite):
        '''Get the PDE of a single neurite'''
        # Get the X, Y,Z coordinates of the points in each section
        points = neurite.points[:, :3]
        return morphmath.principal_direction_extent(points)[direction]

    return map(_pde, iter_neurites(neurites, filt=is_type(neurite_type)))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Morphometrics functions for neurons or neuron populations'''

import math
import numpy as np
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as is_type
from neurom.core.dataformat import COLS
from neurom import morphmath


def neuron_population(nrns):
    '''Makes sure `nrns` behaves like a neuron population'''
    return nrns.neurons if hasattr(nrns, 'neurons') else (nrns,)


def soma_surface_area(nrn, neurite_type=NeuriteType.soma):
    '''Get the surface area of a neuron's soma.

    Note:
        The surface area is calculated by assuming the soma is spherical.
    '''
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    return 4 * math.pi * nrn.soma.radius ** 2


def soma_surface_areas(nrn_pop, neurite_type=NeuriteType.soma):
    '''Get the surface areas of the somata in a population of neurons

    Note:
        The surface area is calculated by assuming the soma is spherical.
    Note:
        If a single neuron is passed, a single element list with the surface
        area of its soma member is returned.
    '''
    nrns = neuron_population(nrn_pop)
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    return [soma_surface_area(n) for n in nrns]


def soma_radii(nrn_pop, neurite_type=NeuriteType.soma):
    ''' Get the radii of the somata of a population of neurons

    Note:
        If a single neuron is passed, a single element list with the
        radius of its soma member is returned.
    '''
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    nrns = neuron_population(nrn_pop)
    return [n.soma.radius for n in nrns]


def trunk_section_lengths(nrn, neurite_type=NeuriteType.all):
    '''list of lengths of trunk sections of neurites in a neuron'''
    neurite_filter = is_type(neurite_type)
    return [morphmath.section_length(s.root_node.points)
            for s in nrn.neurites if neurite_filter(s)]


def trunk_origin_radii(nrn, neurite_type=NeuriteType.all):
    '''radii of the trunk sections of neurites in a neuron'''
    neurite_filter = is_type(neurite_type)
    return [s.root_node.points[0][COLS.R] for s in nrn.neurites if neurite_filter(s)]


def trunk_origin_azimuths(nrn, neurite_type=NeuriteType.all):
    '''Get a list of all the trunk origin azimuths of a neuron or population

    The azimuth is defined as Angle between x-axis and the vector
    defined by (initial tree point - soma center) on the x-z plane.

    The range of the azimuth angle [-pi, pi] radians
    '''
    neurite_filter = is_type(neurite_type)
    nrns = neuron_population(nrn)

    def _azimuth(section, soma):
        '''Azimuth of a section'''
        vector = morphmath.vector(section[0], soma.center)
        return np.arctan2(vector[COLS.Z], vector[COLS.X])

    return [_azimuth(s.root_node.points, n.soma)
            for n in nrns
            for s in n.neurites if neurite_filter(s)]


def trunk_origin_elevations(nrn, neurite_type=NeuriteType.all):
    '''Get a list of all the trunk origin elevations of a neuron or population

    The elevation is defined as the angle between x-axis and the
    vector defined by (initial tree point - soma center)
    on the x-y half-plane.

    The range of the elevation angle [-pi/2, pi/2] radians
    '''
    neurite_filter = is_type(neurite_type)
    nrns = neuron_population(nrn)

    def _elevation(section, soma):
        '''Elevation of a section'''
        vector = morphmath.vector(section[0], soma.center)
        norm_vector = np.linalg.norm(vector)

        if norm_vector >= np.finfo(type(norm_vector)).eps:
            return np.arcsin(vector[COLS.Y] / norm_vector)
        else:
            raise ValueError("Norm of vector between soma center and section is almost zero.")

    return [_elevation(s.root_node.points, n.soma)
            for n in nrns
            for s in n.neurites if neurite_filter(s)]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Neuron classes and functions'''

from itertools import chain
from copy import deepcopy
import numpy as np

from . import NeuriteType, Tree
from neurom import morphmath
from neurom.core.dataformat import COLS
from neurom.utils import memoize
from neurom._compat import filter, map, zip


def iter_neurites(obj, mapfun=None, filt=None):
    '''Iterator to a neurite, neuron or neuron population

    Applies optional neurite filter and mapping functions.

    Parameters:
        obj: a neurite, neuron or neuron population.
        mapfun: optional neurite mapping function.
        filt: optional neurite filter function.

    Examples:

        Get the number of points in each neurite in a neuron population

        >>> from neurom.core import iter_neurites
        >>> n_points = [n for n in iter_neurites(pop, lambda x : len(x.points))]

        Get the number of points in each axon in a neuron population

        >>> import neurom as nm
        >>> from neurom.core import iter_neurites
        >>> filter = lambda n : n.type == nm.AXON
        >>> mapping = lambda n : len(n.points)
        >>> n_points = [n for n in iter_neurites(pop, mapping, filter)]

    '''
    neurites = ((obj,) if isinstance(obj, Neurite) else
                obj.neurites if hasattr(obj, 'neurites') else obj)

    neurite_iter = iter(neurites) if filt is None else filter(filt, neurites)
    return neurite_iter if mapfun is None else map(mapfun, neurite_iter)


def iter_sections(neurites, iterator_type=Tree.ipreorder, neurite_filter=None):
    '''Iterator to the sections in a neurite, neuron or neuron population.

    Parameters:
        neurites: neuron, population, neurite, or iterable containing neurite objects
        iterator_type: type of the iteration (ipreorder, iupstream, ibifurcation_point)
        neurite_filter: optional top level filter on properties of neurite neurite objects.

    Examples:

        Get the number of points in each section of all the axons in a neuron population

        >>> import neurom as nm
        >>> from neurom.core import ites_sections
        >>> filter = lambda n : n.type == nm.AXON
        >>> n_points = [len(s.points) for s in iter_sections(pop,  neurite_filter=filter)]

    '''
    return chain.from_iterable(iterator_type(neurite.root_node)
                               for neurite in iter_neurites(neurites, filt=neurite_filter))


def iter_segments(obj, neurite_filter=None):
    '''Return an iterator to the segments in a collection of neurites

    Parameters:
        obj: neuron, population, neurite, section, or iterable containing neurite objects
        neurite_filter: optional top level filter on properties of neurite neurite objects

    Note:
        This is a convenience function provideded for generic access to
        neuron segments. It may have a performance overhead WRT custom-made
        segment analysis functions that leverage numpy and section-wise iteration.
    '''
    sections = iter((obj,) if isinstance(obj, Section) else
                    iter_sections(obj, neurite_filter=neurite_filter))

    return chain.from_iterable(zip(sec.points[:-1], sec.points[1:])
                               for sec in sections)


class Section(Tree):
    '''Class representing a neurite section'''
    def __init__(self, points, section_id=None, section_type=NeuriteType.undefined):
        super(Section, self).__init__()
        self.id = section_id
        self.points = points
        self.type = section_type

    @property
    @memoize
    def length(self):
        '''Return the path length of this section.'''
        return morphmath.section_length(self.points)

    @property
    @memoize
    def area(self):
        '''Return the surface area of this section.

        The area is calculated from the segments, as defined by this
        section's points
        '''
        return sum(morphmath.segment_area(s) for s in iter_segments(self))

    @property
    @memoize
    def volume(self):
        '''Return the volume of this section.

        The volume is calculated from the segments, as defined by this
        section's points
        '''
        return sum(morphmath.segment_volume(s) for s in iter_segments(self))

    def __str__(self):
        return 'Section(id = %s, points=%s) <parent: %s, nchildren: %d>' % \
            (self.id, self.points, self.parent, len(self.children))


class Neurite(object):
    '''Class representing a neurite tree'''
    def __init__(self, root_node):
        self.root_node = root_node
        self.type = root_node.type if hasattr(root_node, 'type') else NeuriteType.undefined

    @property
    @memoize
    def points(self):
        '''Return unordered array with all the points in this neurite'''
        # add all points in a section except the first one, which is a duplicate
        _pts = [v for s in self.root_node.ipreorder()
                for v in s.points[1:, COLS.XYZR]]
        # except for the very first point, which is not a duplicate
        _pts.insert(0, self.root_node.points[0][COLS.XYZR])
        return np.array(_pts)

    @property
    @memoize
    def length(self):
        '''Return the total length of this neurite.

        The length is defined as the sum of lengths of the sections.
        '''
        return sum(s.length for s in self.iter_sections())

    @property
    @memoize
    def area(self):
        '''Return the surface area of this neurite.

        The area is defined as the sum of area of the sections.
        '''
        return sum(s.area for s in self.iter_sections())

    @property
    @memoize
    def volume(self):
        '''Return the volume of this neurite.

        The volume is defined as the sum of volumes of the sections.
        '''
        return sum(s.volume for s in self.iter_sections())

    def transform(self, trans):
        '''Return a copy of this neurite with a 3D transformation applied'''
        clone = deepcopy(self)
        for n in clone.iter_sections():
            n.points[:, 0:3] = trans(n.points[:, 0:3])

        return clone

    def iter_sections(self, order=Tree.ipreorder):
        '''iteration over section nodes'''
        return iter_sections(self, iterator_type=order)

    def __deepcopy__(self, memo):
        '''Deep copy of neurite object'''
        return Neurite(deepcopy(self.root_node, memo))

    def __nonzero__(self):
        return bool(self.root_node)

    __bool__ = __nonzero__

    def __str__(self):
        return 'Neurite <type: %s>' % self.type


class Neuron(object):
    '''Class representing a simple neuron'''
    def __init__(self, soma=None, neurites=None, sections=None):
        self.soma = soma
        self.neurites = neurites
        self.sections = sections

    def __str__(self):
        return 'Neuron <soma: %s, nneurites: %d>' % \
            (self.soma, len(self.neurites))



# Copyright (c) 2016, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Reader for Neurolucida .ASC files, v3, reversed engineered from looking at output from
Neuroludica
'''

import logging
import warnings
from neurom._compat import StringType

import numpy as np

from neurom.core.dataformat import COLS, POINT_TYPE
from .datawrapper import DataWrapper


WANTED_SECTIONS = {
    'CellBody': POINT_TYPE.SOMA,
    'Axon': POINT_TYPE.AXON,
    'Dendrite': POINT_TYPE.BASAL_DENDRITE,
    'Apical': POINT_TYPE.APICAL_DENDRITE,
}
UNWANTED_SECTION_NAMES = [
    # Meta-data?
    'Closed', 'Color', 'FillDensity', 'GUID', 'ImageCoords', 'MBFObjectType',
    'Marker', 'Name', 'Resolution', 'Set',
    # Marker names?
    'Asterisk', 'Cross', 'Dot', 'DoubleCircle', 'FilledCircle', 'FilledDownTriangle',
    'FilledSquare', 'FilledStar', 'FilledUpTriangle', 'FilledUpTriangle', 'Flower',
    'Flower2', 'OpenCircle', 'OpenDiamond', 'OpenDownTriangle', 'OpenSquare', 'OpenStar',
    'OpenUpTriangle', 'Plus', 'ShadedStar', 'Splat', 'TriStar',
]
UNWANTED_SECTIONS = dict([(name, True) for name in UNWANTED_SECTION_NAMES])
L = logging.getLogger(__name__)


def _match_section(section, match):
    '''checks whether the `type` of section is in the `match` dictionary

    Works around the unknown ordering of s-expressions in each section.
    For instance, the `type` is the 3-rd one in for CellBodies
        ("CellBody"
         (Color Yellow)
         (CellBody)
         (Set "cell10")
        )

    Returns:
        value associated with match[section_type], None if no match
    '''
    # TODO: rewrite this so it is more clear, and handles sets & dictionaries for matching
    for i in range(5):
        if i >= len(section):
            return None
        elif isinstance(section[i], StringType) and section[i] in match:
            return match[section[i]]
    return None


def _get_tokens(morph_fd):
    '''split a file-like into tokens: split on whitespace

    Note: this also strips newlines and comments
    '''
    for line in morph_fd.readlines():
        line = line.rstrip()   # remove \r\n
        line = line.split(';', 1)[0]  # strip comments
        squash_token = []  # quoted strings get squashed into one token
        for token in line.replace('(', ' ( ').replace(')', ' ) ').split():
            if squash_token:
                squash_token.append(token)
                if token.endswith('"'):
                    token = ' '.join(squash_token)
                    squash_token = []
                    yield token
            elif token.startswith('"') and not token.endswith('"'):
                squash_token.append(token)
            else:
                yield token


def _parse_section(token_iter):
    '''take a stream of tokens, and create the tree structure that is defined
    by the s-expressions
    '''
    sexp = []
    for token in token_iter:
        if '(' == token:
            new_sexp = _parse_section(token_iter)
            if not _match_section(new_sexp, UNWANTED_SECTIONS):
                sexp.append(new_sexp)
        elif ')' == token:
            return sexp
        else:
            sexp.append(token)
    return sexp


def _parse_sections(morph_fd):
    '''returns array of all the sections that exist

    The format is nested lists that correspond to the s-expressions
    '''
    sections = []
    token_iter = _get_tokens(morph_fd)
    for token in token_iter:
        if '(' == token:  # find top-level sections
            section = _parse_section(token_iter)
            if not _match_section(section, UNWANTED_SECTIONS):
                sections.append(section)
    return sections


def _flatten_subsection(subsection, _type, offset, parent):
    '''Flatten a subsection from its nested version

    Args:
        subsection: Nested subsection as produced by _parse_section, except one level in
        _type: type of section, ie: AXON, etc
        parent: first element has this as it's parent
        offset: position in the final array of the first element

    Returns:
        Generator of values corresponding to [X, Y, Z, R, TYPE, ID, PARENT_ID]
    '''
    for row in subsection:
        # TODO: Figure out what these correspond to in neurolucida
        if row in ('Low', 'Generated', 'High', ):
            continue
        elif isinstance(row[0], StringType):
            if len(row) in (4, 5, ):
                if 5 == len(row):
                    assert 'S' == row[4][0], \
                        'Only known usage of a fifth member is Sn, found: %s' % row[4][0]
                yield (float(row[0]), float(row[1]), float(row[2]), float(row[3]) / 2.,
                       _type, offset, parent)
                parent = offset
                offset += 1
        elif isinstance(row[0], list):
            split_parent = offset - 1
            start_offset = 0

            slices = []
            start = 0
            for i, value in enumerate(row):
                if '|' == value:
                    slices.append(slice(start + start_offset, i))
                    start = i + 1
            slices.append(slice(start + start_offset, len(row)))

            for split_slice in slices:
                for _row in _flatten_subsection(row[split_slice], _type, offset,
                                                split_parent):
                    offset += 1
                    yield _row


def _extract_section(section):
    '''Find top level sections, and get their flat contents, and append them all

    Returns a numpy array with the row format:
        [X, Y, Z, R, TYPE, ID, PARENT_ID]

    Note: PARENT_ID starts at -1 for soma and 0 for neurites
    '''
    # sections with only one element will be skipped,
    if 1 == len(section):
        assert 'Sections' == section[0], \
            ('Only known usage of a single Section content is "Sections", found %s' %
             section[0])
        return None

    # try and detect type
    _type = WANTED_SECTIONS.get(section[0][0], None)

    start = 1
    # CellBody often has [['"CellBody"'], ['CellBody'] as its first two elements
    if _type is None:
        _type = WANTED_SECTIONS.get(section[1][0], None)

        if _type is None:  # can't determine the type
            return None
        start = 2

    parent = -1 if _type == POINT_TYPE.SOMA else 0
    subsection_iter = _flatten_subsection(section[start:], _type, offset=0,
                                          parent=parent)

    ret = np.array([row for row in subsection_iter])
    return ret


def _sections_to_raw_data(sections):
    '''convert list of sections into the `raw_data` format used in neurom

    This finds the soma, and attaches the neurites
    '''
    soma = None
    neurites = []
    for section in sections:
        neurite = _extract_section(section)
        if neurite is None:
            continue
        elif neurite[0][COLS.TYPE] == POINT_TYPE.SOMA:
            assert soma is None, 'Multiple somas defined in file'
            soma = neurite
        else:
            neurites.append(neurite)
    assert soma is not None, 'No soma found'

    total_length = len(soma) + sum(len(neurite) for neurite in neurites)
    ret = np.zeros((total_length, 7,), dtype=np.float64)
    pos = len(soma)
    ret[0:pos, :] = soma

    for neurite in neurites:
        end = pos + len(neurite)
        ret[pos:end, :] = neurite
        ret[pos:end, COLS.P] += pos
        ret[pos:end, COLS.ID] += pos
        # TODO: attach the neurite at the closest point on the soma
        ret[pos, COLS.P] = len(soma) - 1
        pos = end

    return ret


def read(morph_file, data_wrapper=DataWrapper):
    '''return a 'raw_data' np.array with the full neuron, and the format of the file
    suitable to be wrapped by DataWrapper
    '''

    msg = ('This is an experimental reader. '
           'There are no guarantees regarding ability to parse '
           'Neurolucida .asc files or correctness of output.')

    warnings.warn(msg)
    L.warning(msg)

    with open(morph_file) as morph_fd:
        sections = _parse_sections(morph_fd)
    raw_data = _sections_to_raw_data(sections)
    return data_wrapper(raw_data, 'NL-ASCII')
